######  This file contains the functions needed from the other files to  #######
## perform the analysis presented in the corresponding paper (Zanella 2014) ####  

###########  Packages needed ############
###
library(rgdal)
library(LearnBayes)
library(slam)
library(MASS)
library(fields)
library(coda)
library(mapdata) 
library(spatstat)

###############################################
###########  FUNCTION FOR 2 COLORS ############
###############################################

#### MH algorithm on the matching/partition space (with different proposals) #####
###
twod_MH_P1=function(weights,start_match=NULL,T,burn_time=0,thinning=1,print_time=FALSE,trunc_tresh=0,ref_match_list=NULL,eval_links_count=FALSE,links_count_ref=NULL){
  # This function runs a MH algorithm on the space of partial matchings of a  bipartite graph.
  # The target measure is the product of the edge weights given by the matrix weights.
  # The proposal used is described in the corresponding paper (Zanella 2014) in equation (P1)
  ###### INITIALIZES STATE AND PARAMETERS FOR THE MCMC
  nx<-dim(weights)[1];ny<-dim(weights)[2]
  ntrunc_mat<-weights>trunc_tresh
  weights[!ntrunc_mat]<-0
  ij_nt<-which(ntrunc_mat)
  if (length(start_match)==0){matching<-matrix(rep(0,ntimes=nx),nrow=nx)
  }else{matching<-start_match}
  ###### INITIALIZES HISTORY VECTORS THAT WILL BE OUTPUT #####
  links_count<-matrix(rep(0,nx*ny),nx,ny)
  n_clustered<-rep(NA,(T+burn_time)%/%thinning)
  acceptance_rates<- rep(NA,(T+burn_time)%/%thinning-1)
  diff_from_ref<-NULL
  if(length(ref_match_list)>0){
    diff_from_ref<-matrix(rep(NA,length(ref_match_list)*((T+burn_time)%/%thinning) ),nrow=length(ref_match_list))
    for(iter in 1:length(ref_match_list)){
      diff_from_ref[iter,1]<-sum(matching!=ref_match_list[[iter]])
    }
  }
  n_clustered[1] <- sum(matching!=0)
  D_vec<-NULL
  if(length(links_count_ref)>0){
    D_vec<- rep(NA,(T-1)%/%thinning)
  }
  #### MCMC ALGORITHM #####
  for(t in 1:(T+burn_time-1)){
    ########### PROPOSING AN EDGE ACCORDING TO P1 #####
    ij<-ij_nt[sample(length(ij_nt),1)]
    j<-((ij+nx-1)%/%nx)
    i<-ij-nx*(j-1)
    ########### EVALUATES THE ACCEPTING PROBABILITY #########
    if (matching[i]==j){#CASE 1: x&y ARE MATCHED TOGHETER: REMOVE THE LINK
      accepting_prob<-1/weights[i,j]
    }else{
      if (matching[i]==0 & sum(matching==j)==0){#CASE 2: x&y ARE FREE: ADD THE LINK
        accepting_prob<-weights[i,j]
      }else{
        if (matching[i]==0 & sum(matching==j)==1){#CASE 3: x IS FREE y IS BUSY: 'PIVOT' ON y
          accepting_prob<-weights[i,j]/weights[which(matching==j),j]
        }else{
          if (matching[i]!=0 & sum(matching==j)==0){#CASE 4: x IS BUSY y IS FREE: 'PIVOT' ON x
            accepting_prob<-weights[i,j]/weights[i,matching[i]]
          }else{
            if (matching[i]!=0 & sum(matching==j)==1){#CASE 5: x&y ARE BOTH BUSY: SWITCH THE LINKS 
              accepting_prob<-weights[i,j]*weights[which(matching==j),matching[i]]/
                (weights[i,matching[i]]*weights[which(matching==j),j])
            }
          }
        }
      }
    }
    #### ACCEPT THE MOVE WITH THE ACCEPTING PROBABILITY
    if (runif(1)<accepting_prob){
      matching<-matching_proposed(matching=matching,i=i,j=j)
    }
    ### UPDATE HISTORY VECTORS THAT WILL BE OUTPUT ####
    if(t%%thinning==0){
      n_clustered[t%/%thinning+1]<-sum(matching!=0)
      acceptance_rates[t%/%thinning]<-accepting_prob
      if(t>burn_time & eval_links_count){
        links_count<-links_count+(matrix(rep(matching,ny),nrow=nx,ncol=ny,byrow=FALSE)==matrix(rep(c(1:ny),nx),nrow=nx,ncol=ny,byrow=TRUE))
        if(length(links_count_ref)>0){
          D_vec[(t-burn_time)%/%thinning]<-max(abs( links_count_ref-links_count/( (t-burn_time) %/%thinning) ))  
        }
      }
      if(length(ref_match_list)>0){
        for(iter in 1:length(ref_match_list)){
          diff_from_ref[iter,t%/%thinning+1]<-sum(matching!=ref_match_list[[iter]])
        }
      }
    }
    if(print_time){if(t%%100==0){print(paste('time=',t))}}
  }
  links_count<-links_count/((T)%/%thinning)
  return(list(links_count=links_count,matching=matching,n_clustered=n_clustered,acceptance_rates=acceptance_rates,diff_from_ref=diff_from_ref,D_vec=D_vec))
}
twod_MH_P2=function(weights,start_match=NULL,T,burn_time=0,thinning=1,print_time=FALSE,ref_match_list=NULL,eval_links_count=FALSE,links_count_ref=NULL){
  # This function runs a MH algorithm on the space of partial matchings of a  bipartite graph.
  # The target measure is the product of the edge weights given by the matrix weights.
  # The proposal used is described in the corresponding paper (Zanella 2014) in equation (P2)
  ###### INITIALIZES STATE AND PARAMETERS FOR THE MCMC
  nx<-dim(weights)[1];ny<-dim(weights)[2]
  if (length(start_match)==0){matching<-matrix(rep(0,ntimes=nx),nrow=nx)
  }else{matching<-start_match}
  prop_mat<-build_post_mat(weights=weights,matching=matching)
  norm_const<-sum(prop_mat)
  marginal_i<-rowSums(prop_mat)
  ###### INITIALIZES HISTORY VECTORS THAT WILL BE OUTPUT #####
  links_count<-matrix(rep(0,nx*ny),nx,ny)
  n_clustered<-rep(NA,(T+burn_time)%/%thinning)
  acceptance_rates<- rep(NA,(T+burn_time)%/%thinning-1)
  diff_from_ref<-NULL
  if(length(ref_match_list)>0){
    diff_from_ref<-matrix(rep(NA,length(ref_match_list)*((T+burn_time)%/%thinning) ),nrow=length(ref_match_list))
    for(iter in 1:length(ref_match_list)){
      diff_from_ref[iter,1]<-sum(matching!=ref_match_list[[iter]])
    }
  }
  n_clustered[1] <- sum(matching!=0)
  D_vec<-NULL
  if(length(links_count_ref)>0){
    D_vec<- rep(NA,(T-1)%/%thinning)
  }
  ##### MCMC ALGORITHM #######
  for(t in 1:(T+burn_time-1)){
    ########### PROPOSING AN EDGE ACCORDING TO P2 #####
    cum_probs<-cumsum(marginal_i)
    i<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    cum_probs<-cumsum(prop_mat[i,])
    j<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    ########### EVALUATES THE ACCEPTING PROBABILITY AND ACCEPTS THE MOVE #########
    new_matching<-matching_proposed(matching=matching,i=i,j=j)
    new_prop_row_i<-update_row(weights=weights,i=i,current_matching=new_matching)
    new_prop_col_j<-update_column(weights=weights,j=j,current_matching=new_matching)
    if (matching[i]==j | (matching[i]==0 & sum(matching==j)==0)){#ADDITION OR DELETION MOVE
      delta_i<-sum(new_prop_row_i[-j])-sum(prop_mat[i,-j])
      new_norm_const<-norm_const+delta_i+sum(new_prop_col_j)-sum(prop_mat[,j])
      ap<-1/prop_mat[i,j]* norm_const/new_norm_const
      if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
        marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]
        marginal_i[i]<-marginal_i[i]+delta_i
        prop_mat[i,]<-new_prop_row_i
        prop_mat[,j]<-new_prop_col_j
        norm_const<-new_norm_const
        matching<-new_matching
      }
    }else{
      if (matching[i]==0 & sum(matching==j)==1){#SWITCH MOVE ('PIVOT' ON y)
        i_prime<-which(matching==j)
        new_prop_row_i_prime<-(update_row(weights=weights,i=i_prime,current_matching=new_matching))
        delta_i<-sum(new_prop_row_i[-j])-sum(prop_mat[i,-j])
        delta_i_prime<-sum(new_prop_row_i_prime[-j])-sum(prop_mat[i_prime,-j])
        new_norm_const<-norm_const+delta_i+delta_i_prime+sum(new_prop_col_j)-sum(prop_mat[,j])
        ap<-1/prop_mat[i,j]*norm_const/new_norm_const
        if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
          marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]
          marginal_i[i]<-marginal_i[i]+delta_i
          marginal_i[i_prime]<-marginal_i[i_prime]+delta_i_prime
          prop_mat[i,]<-new_prop_row_i
          prop_mat[i_prime,]<-new_prop_row_i_prime
          prop_mat[,j]<-new_prop_col_j
          norm_const<-new_norm_const
          matching<-new_matching
        }
      }else{
        if (matching[i]!=0 & sum(matching==j)==0){#SWITCH MOVE ('PIVOT' ON X)
          delta_i<-sum(new_prop_row_i[-c(j,matching[i])])-sum(prop_mat[i,-c(j,matching[i])])
          new_prop_col_j_prime<-(update_column(weights=weights,j=matching[i],current_matching=new_matching))
          new_norm_const<-norm_const+delta_i+sum(new_prop_col_j)-sum(prop_mat[,j])+sum(new_prop_col_j_prime)-sum(prop_mat[,matching[i]])
          ap<-1/prop_mat[i,j]*norm_const/new_norm_const
          if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
            marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]+new_prop_col_j_prime-prop_mat[,matching[i]]
            marginal_i[i]<-marginal_i[i]+delta_i
            prop_mat[i,]<-new_prop_row_i
            prop_mat[,j]<-new_prop_col_j
            prop_mat[,matching[i]]<-new_prop_col_j_prime
            norm_const<-new_norm_const
            matching<-new_matching
          }
        }else{
          if (matching[i]!=0 & sum(matching==j)==1){#DOUBLE-SWITCH MOVE
            delta_i<-sum(new_prop_row_i[-c(j,matching[i])])-sum(prop_mat[i,-c(j,matching[i])])
            i_prime<-which(matching==j)
            new_prop_row_i_prime<-(update_row(weights=weights,i=i_prime,current_matching=new_matching))
            delta_i_prime<-sum(new_prop_row_i_prime[-c(j,matching[i])])-sum(prop_mat[i_prime,-c(j,matching[i])])
            new_prop_col_j_prime<-(update_column(weights=weights,j=matching[i],current_matching=new_matching))
            new_norm_const<-norm_const+delta_i+delta_i_prime+sum(new_prop_col_j)-sum(prop_mat[,j])+sum(new_prop_col_j_prime)-sum(prop_mat[,matching[i]])
            ap<-1/prop_mat[i,j]*norm_const/new_norm_const
            if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
              marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]+new_prop_col_j_prime-prop_mat[,matching[i]]
              marginal_i[i]<-marginal_i[i]+delta_i
              marginal_i[i_prime]<-marginal_i[i_prime]+delta_i_prime
              prop_mat[i,]<-new_prop_row_i
              prop_mat[i_prime,]<-new_prop_row_i_prime
              prop_mat[,j]<-new_prop_col_j
              prop_mat[,matching[i]]<-new_prop_col_j_prime
              norm_const<-new_norm_const
              matching<-new_matching
            }  
          }
        }
      }
    }
    ### UPDATE HISTORY VECTORS THAT WILL BE OUTPUT ####
    if(t%%thinning==0){
      n_clustered[t%/%thinning+1]<-sum(matching!=0)
      acceptance_rates[t%/%thinning]<-ap
      if(t>burn_time & eval_links_count){
        links_count<-links_count+(matrix(rep(matching,ny),nrow=nx,ncol=ny,byrow=FALSE)==matrix(rep(c(1:ny),nx),nrow=nx,ncol=ny,byrow=TRUE))
        if(length(links_count_ref)>0){
          D_vec[(t-burn_time)%/%thinning]<-max(abs( links_count_ref-links_count/( (t-burn_time) %/%thinning) ))  
        }
      }
      if(length(ref_match_list)>0){
        for(iter in 1:length(ref_match_list)){
          diff_from_ref[iter,t%/%thinning+1]<-sum(matching!=ref_match_list[[iter]])
        }
      }
    }
    if(print_time){if(t%%100==0){print(paste('time=',t))}}
  }
  links_count<-links_count/((T+burn_time)%/%thinning)
  return(list(links_count=links_count,matching=matching,n_clustered=n_clustered,acceptance_rates=acceptance_rates,diff_from_ref=diff_from_ref,D_vec=D_vec))
}
twod_MH_P3=function(weights,start_match=NULL,T,burn_time=0,thinning=1,print_time=FALSE,ref_match_list=NULL,eval_links_count=FALSE,links_count_ref=NULL){
  # This function runs a MH algorithm on the space of partial matchings of a  bipartite graph.
  # The target measure is the product of the edge weights given by the matrix weights.
  # The proposal used is described in the corresponding paper (Zanella 2014) in equation (P3)
  ###### INITIALIZES STATE AND PARAMETERS FOR THE MCMC
  nx<-dim(weights)[1];ny<-dim(weights)[2]
  if (length(start_match)==0){matching<-matrix(rep(0,ntimes=nx),nrow=nx)
  }else{matching<-start_match}
  prop_mat<-q_function(build_post_mat(weights=weights,matching=matching))
  norm_const<-sum(prop_mat)
  marginal_i<-rowSums(prop_mat)
  ###### INITIALIZES HISTORY VECTORS THAT WILL BE OUTPUT #####
  links_count<-matrix(rep(0,nx*ny),nx,ny)
  n_clustered<-rep(NA,(T+burn_time)%/%thinning)
  acceptance_rates<- rep(NA,(T+burn_time)%/%thinning-1)
  diff_from_ref<-NULL
  if(length(ref_match_list)>0){
    diff_from_ref<-matrix(rep(NA,length(ref_match_list)*((T+burn_time)%/%thinning) ),nrow=length(ref_match_list))
    for(iter in 1:length(ref_match_list)){
      diff_from_ref[iter,1]<-sum(matching!=ref_match_list[[iter]])
    }
  }
  n_clustered[1] <- sum(matching!=0)
  D_vec<-NULL
  if(length(links_count_ref)>0){
    D_vec<- rep(NA,(T-1)%/%thinning)
  }
  ##### MCMC ALGORITHM #######
  for(t in 1:(T+burn_time-1)){
    ########### PROPOSING AN EDGE ACCORDING TO P3 #####
    cum_probs<-cumsum(marginal_i)
    i<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    cum_probs<-cumsum(prop_mat[i,])
    j<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]    
    ########### EVALUATES THE ACCEPTING PROBABILITY AND ACCEPTS THE MOVE #########
    new_matching<-matching_proposed(matching=matching,i=i,j=j)
    new_prop_row_i<-q_function(update_row(weights=weights,i=i,current_matching=new_matching))
    new_prop_col_j<-q_function(update_column(weights=weights,j=j,current_matching=new_matching))
    if (matching[i]==j | (matching[i]==0 & sum(matching==j)==0)){#ADDITION OR DELETION MOVE
      delta_i<-sum(new_prop_row_i[-j])-sum(prop_mat[i,-j])
      new_norm_const<-norm_const+delta_i+sum(new_prop_col_j)-sum(prop_mat[,j])
      ap<-norm_const/new_norm_const
      if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
        marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]
        marginal_i[i]<-marginal_i[i]+delta_i
        prop_mat[i,]<-new_prop_row_i
        prop_mat[,j]<-new_prop_col_j
        norm_const<-new_norm_const
        matching<-new_matching
      }
    }else{
      if (matching[i]==0 & sum(matching==j)==1){#SWITCH MOVE ('PIVOT' ON y)
        i_prime<-which(matching==j)
        new_prop_row_i_prime<-q_function(update_row(weights=weights,i=i_prime,current_matching=new_matching))
        delta_i<-sum(new_prop_row_i[-j])-sum(prop_mat[i,-j])
        delta_i_prime<-sum(new_prop_row_i_prime[-j])-sum(prop_mat[i_prime,-j])
        new_norm_const<-norm_const+delta_i+delta_i_prime+sum(new_prop_col_j)-sum(prop_mat[,j])
        ap<-norm_const/new_norm_const
        if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
          marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]
          marginal_i[i]<-marginal_i[i]+delta_i
          marginal_i[i_prime]<-marginal_i[i_prime]+delta_i_prime
          prop_mat[i,]<-new_prop_row_i
          prop_mat[i_prime,]<-new_prop_row_i_prime
          prop_mat[,j]<-new_prop_col_j
          norm_const<-new_norm_const
          matching<-new_matching
        }
      }else{
        if (matching[i]!=0 & sum(matching==j)==0){#SWITCH MOVE ('PIVOT' ON X)
          delta_i<-sum(new_prop_row_i[-c(j,matching[i])])-sum(prop_mat[i,-c(j,matching[i])])
          new_prop_col_j_prime<-q_function(update_column(weights=weights,j=matching[i],current_matching=new_matching))
          new_norm_const<-norm_const+delta_i+sum(new_prop_col_j)-sum(prop_mat[,j])+sum(new_prop_col_j_prime)-sum(prop_mat[,matching[i]])
          ap<-norm_const/new_norm_const
          if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
            marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]+new_prop_col_j_prime-prop_mat[,matching[i]]
            marginal_i[i]<-marginal_i[i]+delta_i
            prop_mat[i,]<-new_prop_row_i
            prop_mat[,j]<-new_prop_col_j
            prop_mat[,matching[i]]<-new_prop_col_j_prime
            norm_const<-new_norm_const
            matching<-new_matching
          }
        }else{
          if (matching[i]!=0 & sum(matching==j)==1){#DOUBLE-SWITCH MOVE
            delta_i<-sum(new_prop_row_i[-c(j,matching[i])])-sum(prop_mat[i,-c(j,matching[i])])
            i_prime<-which(matching==j)
            new_prop_row_i_prime<-q_function(update_row(weights=weights,i=i_prime,current_matching=new_matching))
            delta_i_prime<-sum(new_prop_row_i_prime[-c(j,matching[i])])-sum(prop_mat[i_prime,-c(j,matching[i])])
            new_prop_col_j_prime<-q_function(update_column(weights=weights,j=matching[i],current_matching=new_matching))
            new_norm_const<-norm_const+delta_i+delta_i_prime+sum(new_prop_col_j)-sum(prop_mat[,j])+sum(new_prop_col_j_prime)-sum(prop_mat[,matching[i]])
            ap<-norm_const/new_norm_const
            if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
              marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]+new_prop_col_j_prime-prop_mat[,matching[i]]
              marginal_i[i]<-marginal_i[i]+delta_i
              marginal_i[i_prime]<-marginal_i[i_prime]+delta_i_prime
              prop_mat[i,]<-new_prop_row_i
              prop_mat[i_prime,]<-new_prop_row_i_prime
              prop_mat[,j]<-new_prop_col_j
              prop_mat[,matching[i]]<-new_prop_col_j_prime
              norm_const<-new_norm_const
              matching<-new_matching
            }  
          }
        }
      }
    }
    ### UPDATE HISTORY VECTORS THAT WILL BE OUTPUT ####
    if(t%%thinning==0){
      n_clustered[t%/%thinning+1]<-sum(matching!=0)
      acceptance_rates[t%/%thinning]<-ap
      if(t>burn_time & eval_links_count){
        links_count<-links_count+(matrix(rep(matching,ny),nrow=nx,ncol=ny,byrow=FALSE)==matrix(rep(c(1:ny),nx),nrow=nx,ncol=ny,byrow=TRUE))
        if(length(links_count_ref)>0){
          D_vec[(t-burn_time)%/%thinning]<-max(abs( links_count_ref-links_count/( (t-burn_time) %/%thinning) ))  
        }
      }
      if(length(ref_match_list)>0){
        for(iter in 1:length(ref_match_list)){
          diff_from_ref[iter,t%/%thinning+1]<-sum(matching!=ref_match_list[[iter]])
        }
      }
    }
    if(print_time){if(t%%100==0){print(paste('time=',t))}}
  }
  links_count<-links_count/((T+burn_time)%/%thinning)
  return(list(links_count=links_count,matching=matching,n_clustered=n_clustered,acceptance_rates=acceptance_rates,diff_from_ref=diff_from_ref,D_vec=D_vec))
}
twod_MH_P4=function(weights,start_match=NULL,T,burn_time=0,thinning=1,print_time=FALSE,ref_match_list=NULL,eval_links_count=FALSE,links_count_ref=NULL){
  # This function runs a MH algorithm on the space of partial matchings of a  bipartite graph.
  # The target measure is the product of the edge weights given by the matrix weights.
  # The proposal used is described in the corresponding paper (Zanella 2014) in equation (P4)
  ###### INITIALIZES STATE AND PARAMETERS FOR THE MCMC
  nx<-dim(weights)[1];ny<-dim(weights)[2]
  if (length(start_match)==0){matching<-matrix(rep(0,ntimes=nx),nrow=nx)
  }else{matching<-start_match}
  q_add_mat<-build_q_add_mat(weights)
  q_rem_mat<-build_q_rem_mat(weights)
  prop_mat<-q_add_mat
  links_indices<-matrix(c(which(matching>0),matching[matching>0]),ncol=2)
  prop_mat[links_indices]<-q_rem_mat[links_indices]
  marginal_i<-apply(prop_mat,1,sum)
  norm_const<-sum(marginal_i)
  ###### INITIALIZES HISTORY VECTORS THAT WILL BE OUTPUT #####
  links_count<-matrix(rep(0,nx*ny),nx,ny)
  n_clustered<-rep(NA,(T+burn_time)%/%thinning)
  acceptance_rates<- rep(NA,(T+burn_time)%/%thinning-1)
  diff_from_ref<-NULL
  if(length(ref_match_list)>0){
    diff_from_ref<-matrix(rep(NA,length(ref_match_list)*((T+burn_time)%/%thinning) ),nrow=length(ref_match_list))
    for(iter in 1:length(ref_match_list)){
      diff_from_ref[iter,1]<-sum(matching!=ref_match_list[[iter]])
    }
  }
  n_clustered[1] <- sum(matching!=0)
  D_vec<-NULL
  if(length(links_count_ref)>0){
    D_vec<- rep(NA,(T-1)%/%thinning)
  }
  #### MCMC ALGORITHM #####
  for(t in 1:(T+burn_time-1)){
    ########### PROPOSING AN EDGE ACCORDING TO P4 #####
    cum_probs<-cumsum(marginal_i)
    i<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    cum_probs<-cumsum(prop_mat[i,])
    j<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    ########### EVALUATES THE ACCEPTING PROBABILITY AND ACCEPTS THE MOVE #########
    if (matching[i]==j){#CASE 1: x&y ARE MATCHED TOGHETER: REMOVE THE LINK
      post_ratio<-1/weights[i,j]
      delta_norm_const<-q_add_mat[i,j]-q_rem_mat[i,j]
      ap<-post_ratio*q_add_mat[i,j]/(norm_const+delta_norm_const)/(q_rem_mat[i,j]/norm_const)
      if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
        prop_mat[i,j]<-q_add_mat[i,j]
        marginal_i[i]<-marginal_i[i]+delta_norm_const
        norm_const<-norm_const+delta_norm_const
        matching<-matching_proposed(matching=matching,i=i,j=j)
      }
    }else{
      if (matching[i]==0 & sum(matching==j)==0){#CASE 2: x&y ARE FREE: ADD THE LINK
        post_ratio<-weights[i,j]
        delta_norm_const<-q_rem_mat[i,j]-q_add_mat[i,j]
        ap<-post_ratio*q_rem_mat[i,j]/(norm_const+delta_norm_const)/(q_add_mat[i,j]/norm_const)
        if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
          prop_mat[i,j]<-q_rem_mat[i,j]
          marginal_i[i]<-marginal_i[i]+delta_norm_const
          norm_const<-norm_const+delta_norm_const
          matching<-matching_proposed(matching=matching,i=i,j=j)
        }        
      }else{
        if (matching[i]==0 & sum(matching==j)==1){#CASE 3: x IS FREE y IS BUSY: 'PIVOT' ON y
          i_prime<-which(matching==j)
          post_ratio<-weights[i,j]/weights[i_prime,j]
          delta_norm_const1<-q_rem_mat[i,j]-q_add_mat[i,j]  
          delta_norm_const2<-q_add_mat[i_prime,j]-q_rem_mat[i_prime,j]
          ap<-post_ratio*q_add_mat[i_prime,j]/(norm_const+delta_norm_const1+delta_norm_const2)/(q_add_mat[i,j]/norm_const)
          if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
            prop_mat[i,j]<-q_rem_mat[i,j]
            prop_mat[i_prime,j]<-q_add_mat[i_prime,j]
            marginal_i[i]<-marginal_i[i]+delta_norm_const1   
            marginal_i[i_prime]<-marginal_i[i_prime]+delta_norm_const2
            norm_const<-norm_const+delta_norm_const1+delta_norm_const2
            matching<-matching_proposed(matching=matching,i=i,j=j)
          }
        }else{
          if (matching[i]!=0 & sum(matching==j)==0){#CASE 4: x IS BUSY y IS FREE: 'PIVOT' ON x
            post_ratio<-weights[i,j]/weights[i,matching[i]]
            delta_norm_const<-q_rem_mat[i,j]-q_add_mat[i,j]+q_add_mat[i,matching[i]]-q_rem_mat[i,matching[i]] 
            ap<-post_ratio*q_add_mat[i,matching[i]]/(norm_const+delta_norm_const)/(q_add_mat[i,j]/norm_const)
            if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
              prop_mat[i,j]<-q_rem_mat[i,j]
              prop_mat[i,matching[i]]<-q_add_mat[i,matching[i]]
              marginal_i[i]<-marginal_i[i]+delta_norm_const
              norm_const<-norm_const+delta_norm_const
              matching<-matching_proposed(matching=matching,i=i,j=j)
            }
          }else{
            if (matching[i]!=0 & sum(matching==j)==1){#CASE 5: x&y ARE BOTH BUSY: SWITCH THE LINKS 
              i_prime<-which(matching==j)
              post_ratio<-weights[i,j]*weights[i_prime,matching[i]]/
                (weights[i,matching[i]]*weights[i_prime,j])
              delta_norm_const1<-q_rem_mat[i,j]-q_add_mat[i,j]+q_add_mat[i,matching[i]]-q_rem_mat[i,matching[i]]
              delta_norm_const2<-q_rem_mat[i_prime,matching[i]]-q_add_mat[i_prime,matching[i]]+q_add_mat[i_prime,j]-q_rem_mat[i_prime,j]
              ap<-post_ratio*(q_add_mat[i_prime,j]+q_add_mat[i,matching[i]])/(norm_const+delta_norm_const1+delta_norm_const2)/((q_add_mat[i,j]+q_add_mat[i_prime,matching[i]])/norm_const)
              if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
                prop_mat[i,j]<-q_rem_mat[i,j]
                prop_mat[i,matching[i]]<-q_add_mat[i,matching[i]]
                prop_mat[i_prime,matching[i]]<-q_rem_mat[i_prime,matching[i]]
                prop_mat[i_prime,j]<-q_add_mat[i_prime,j]
                marginal_i[i]<-marginal_i[i]+delta_norm_const1
                marginal_i[i_prime]<-marginal_i[i_prime]+delta_norm_const2                
                norm_const<-norm_const+delta_norm_const1+delta_norm_const2
                matching<-matching_proposed(matching=matching,i=i,j=j)
              }  
            }
          }
        }
      }
    }
    ### UPDATE HISTORY VECTORS THAT WILL BE OUTPUT ####
    if(t%%thinning==0){
      n_clustered[t%/%thinning+1]<-sum(matching!=0)
      acceptance_rates[t%/%thinning]<-ap
      if(t>burn_time & eval_links_count){
        links_mat<-matrix(rep(matching,ny),nrow=nx,ncol=ny,byrow=FALSE)==matrix(rep(c(1:ny),nx),nrow=nx,ncol=ny,byrow=TRUE)
        links_count<-links_count+links_mat
        if(length(links_count_ref)>0){
          D_vec[(t-burn_time)%/%thinning]<-max(abs( links_count_ref-links_count/( (t-burn_time) %/%thinning) ))  
        }
      }
      if(length(ref_match_list)>0){
        for(iter in 1:length(ref_match_list)){
          diff_from_ref[iter,t%/%thinning+1]<-sum(matching!=ref_match_list[[iter]])
        }
      }
    }
    if(print_time){if(t%%100==0){print(paste('time=',t))}}
  }
  links_count<-links_count/(T%/%thinning)
  return(list(links_count=links_count,matching=matching,n_clustered=n_clustered,acceptance_rates=acceptance_rates,diff_from_ref=diff_from_ref,D_vec=D_vec
  ))
}

#### AUXILIARY FUNCTIONS FOR THE Metropolis-Hastings algorithm #############
###
matching_proposed<-function(matching,i,j){
  matching_proposed<-matching
  if (matching[i]==j){#CASE 1: x&y ARE MATCHED TOGHETER: REMOVE THE LINK
    matching_proposed[i]=0 
  }
  else{
    if (matching[i]==0 & sum(matching==j)==0){#CASE 2: x&y ARE FREE: ADD THE LINK
      matching_proposed[i]=j
    }
    else{
      if (matching[i]==0 & sum(matching==j)==1){#CASE 3: x IS FREE y IS BUSY: 'PIVOT' ON y
        matching_proposed[i]=j
        matching_proposed[which(matching==j)]=0
      }
      else{
        if (matching[i]!=0 & sum(matching==j)==0){#CASE 4: x IS BUSY y IS FREE: 'PIVOT' ON x
          matching_proposed[i]=j
        }
        else{
          if (matching[i]!=0 & sum(matching==j)==1){#CASE 5: x&y ARE BOTH BUSY: SWITCH THE LINKS 
            matching_proposed[i]=j
            matching_proposed[which(matching==j)]=matching[i]
          }
        }
      }
    }
  }
  return(matching_proposed)
}
q_function<-function(post_mat){
  return(post_mat/(1+post_mat))
}
build_post_mat<-function(weights,matching){
  #INITIALIZATION
  nb<-sum(matching!=0);nx<-dim(weights)[1];ny<-dim(weights)[2]
  nf_x<-nx-nb;nf_y<-ny-nb;N<-nb+nf_x+nf_y
  links_indices<-matrix(c(which(matching>0),matching[matching>0]),nrow=nb,ncol=2)
  #BUILD PROPOSAL MATRIX
  posterior_matrix<-matrix(rep(0,times=nx*ny),nrow=nx,ncol=ny)
  #PROPOSAL FOR SWITCH MOVES
  if(nb>1){    
    old_weights<-matrix(rep(weights[links_indices],times=nb),nrow=nb,ncol=nb)
    posterior_matrix[matching!=0,matching]<-
      weights[matching!=0,matching]*t(weights[matching!=0,matching])/
      (old_weights*t(old_weights))
  }
  #PROPOSAL FOR REMOVE-LINKS MOVES
  if(nb!=0){  
    posterior_matrix[links_indices]<-1/weights[links_indices]
  }
  #PROPOSAL FOR ADD-LINKS MOVES
  if(nf_x*nf_y>0){
    posterior_matrix[matching==0,-c(matching,(ny+2))]<-
      weights[matching==0,-c(matching,(ny+2))]
  }
  #PROPOSAL FOR x-PIVOTAL MOVES
  if(nb*nf_y>0){
    old_weights<-matrix(rep(weights[links_indices],times=nf_y),nrow=nb,ncol=nf_y)
    posterior_matrix[matching>0,-c(matching,(ny+2))]<-
      weights[matching>0,-c(matching,(ny+2))]/old_weights
  }
  #PROPOSAL FOR y-PIVOTAL MOVES
  if(nb*nf_x>0){
    old_weights<-matrix(rep(weights[links_indices],times=nf_x),nrow=nf_x,ncol=nb,byrow=TRUE)
    posterior_matrix[matching==0,matching]<-
      weights[matching==0,matching]/old_weights
  }
  posterior_matrix[which(posterior_matrix==Inf)]=exp(700)#Machine limit
  return(posterior_matrix)
}
update_row<-function(weights,i,current_matching){
  nb<-sum(current_matching!=0)
  links_indices<-matrix(c(which(current_matching>0),
                          current_matching[current_matching>0]),nrow=nb,ncol=2)
  if (current_matching[i]==0){
    row<-weights[i,]#add moves
    row[links_indices[,2]]<-row[links_indices[,2]]/weights[links_indices]#y_pivot moves
  }else{
    row<-weights[i,]/weights[i,current_matching[i]]#x_pivot moves
    row[links_indices[,2]]<-row[links_indices[,2]]*
      weights[links_indices[,1],current_matching[i]]/
      weights[links_indices]#switch moves
    row[current_matching[i]]<-1/weights[i,current_matching[i]]#remove move
  }
  return(row)
}
update_column<-function(weights,j,current_matching){
  nb<-sum(current_matching!=0)
  links_indices<-matrix(c(which(current_matching>0),
                          current_matching[current_matching>0]),nrow=nb,ncol=2)
  if (any(j==current_matching)){
    i_corr<-which(j==current_matching)
    column<-weights[,j]/weights[i_corr,j]#x_pivot moves
    column[links_indices[,1]]<-column[links_indices[,1]]*
      weights[i_corr,links_indices[,2]]/
      weights[links_indices]#switch moves
    column[i_corr]<-1/weights[i_corr,j]#remove move
  }else{
    column<-weights[,j]#add moves
    column[links_indices[,1]]<-column[links_indices[,1]]/weights[links_indices]#y_pivot moves
  }
  return(column)
}
build_q_add_mat<-function(weights){
  sqrt_weights<-sqrt(weights)
  w_i.<-rep(rowSums(weights),dim(weights)[2])
  w_.j<-rep(colSums(weights),each=dim(weights)[1])
  p_tilde_ij<-(weights-sqrt_weights)/(1+w_.j+w_i.-weights)
  p_tilde_i.<-rep(rowSums(p_tilde_ij),dim(weights)[2])
  p_tilde_.j<-rep(colSums(p_tilde_ij),each=dim(weights)[1])
  q_add_mat<-sqrt_weights*(1-p_tilde_i.+p_tilde_ij)*(1-p_tilde_.j+p_tilde_ij)
  q_add_mat[q_add_mat<0]<-1e-200
  return(q_add_mat)
}
build_q_rem_mat<-function(weights){
  return(1/sqrt(weights))
}
permanentRyser<-function(A){
  #   permanentRyser(A) returns the permanent of square matrix A
  #   INPUT:  A square numeric matrix
  #   OUTPUT: The matrix permanent
  #   Uses the Ryser Formula to calculate the permanent, so it is O((n^2)(2^n))
  #   which is much faster than the naive algorithm O(n!n). The permanent 
  #   of a matrix is defined as the analog of determinant where the signs of 
  #   each term in summation was removed.
  y <- 0
  n<- dim(A)[1]
  n2<- dim(A)[2]
  stopifnot(n==n2)
  # loop over all 2^n submatricies of A
  for (i in 1:(2^n - 1)){#The last 2^n is no rows, skip that
    # use intToBits to index ever possible row-subset of A
    indx <- as.logical(intToBits(i))[1:n]
    y <- y + ((-1)^(sum(indx))) * prod( colSums(matrix(A[indx,],ncol=n)) ) 
  } 
  y <- y * (-1)^n
  return(y)
}

#### FUNCTIONS THAT SPECIFIES PRIOR AND LIKELIHOOD (USE TO ###########
#### EVALUATE THE WEIGHTS MATRIX FOR THE POISSON MODEL)  #############
###
# parameters and function for the intra-cluster dispersion (sigma)
sigma_max<-50*10^3/sqrt(pi)
eval_sum_cl_sq_dist<-function(cluster,n_points,points){
  if (length(cluster)==0){
    sum_cl_sq_dist<-0
  }else{
    sum_cl_sq_dist<-sum(apply(
      cluster,2,intracluster_sqrdist,n_points=n_points,points=points)) 
  }
  return(sum_cl_sq_dist)  
}
# parameters and function for the expected number of clusters (lambda)
scale_lambda<-300
shape_lambda<-1
posterior_scale_lambda<-function(scale_lambda){
  return(scale_lambda/(1+scale_lambda))
}
posterior_shape_lambda<-function(shape_lambda,N_vec){
  return(shape_lambda+sum(N_vec))
}
# parameters and function for the probability of cluster size (p_c)
alpha_p_c_prior<-function(k){
  return(c(rep(1/k,k)))  
}
posterior_alpha_p_c<-function(k,N_vec){
  return(alpha_p_c_prior(k)+N_vec)
}
# function to evaluate weights for the Poisson model
c_vec_poisson<-function(k,p_c,sigma,lambda){
  c_vec<-lambda*p_c/
    (marks_norm_consts[1:k]*c(1:k)*(2*pi*sigma^2)^(c(1:k)-1))
  return(c_vec)
}
evaluate_weights_poisson_2d<-function(xy,sigma,lambda,W,p_c){
  x<-xy$x;y<-xy$y
  nx<-length(x[1,]);ny<-length(y[1,])
  deltas=as.matrix(dist(t(cbind(y,x)),diag=TRUE,upper=TRUE))[(ny+1):(ny+nx),1:(ny)]
  deltas<-as.matrix(deltas)
  coeffs<-lambda*c(p_c[1],1-p_c[1])/
    (c(1,2) * c(2,1) * W* (2*pi*sigma^2)^c(0,1) )
  weights<-coeffs[2]/coeffs[1]^2*exp(-deltas^2/(4*sigma^2))
  return (weights)
}
# function to generate a sample from the Poisson model
poisson_model_synt_sample<-function(W,sigma,lambda,p_c){
  a<-sqrt(W)
  N<-rpois(1,lambda)
  n_clusters<-c(0,0,0)
  x<-c();y<-c()
  for (i in 1:N){
    if (runif(1)<p_c[1]){
      if (runif(1)<0.5){
        x_new<-matrix(runif(2,min=0,max=a),nrow=2,ncol=1)
        y_new<-c()
        n_clusters[1]<-n_clusters[1]+1
      }
      else{
        x_new<-c()
        y_new<-matrix(runif(2,min=0,max=a),nrow=2,ncol=1)
        n_clusters[2]<-n_clusters[2]+1
      }
    }else{
      z=matrix(runif(2,min=0,max=a),nrow=2,ncol=1)#centre
      X1=matrix(rnorm(2,mean=0,sd=sigma),nrow=2,ncol=1) #shift1
      X2=matrix(rnorm(2,mean=0,sd=sigma),nrow=2,ncol=1) #shift2
      x_new<-z+X1-(X1+X2)/2
      y_new<-z+X2-(X1+X2)/2
      n_clusters[3]<-n_clusters[3]+1
    }
    x=cbind(x,x_new);y=cbind(y,y_new)
  }
  n_clusters=c(single_x=n_clusters[1],single_y=n_clusters[2],couples_xy=n_clusters[3])
  return(list(x=x,y=y))
}

##########  AUXILIARY FUNCTIONS FOR PLOTTING #############
###
Plot_matching_gray=function(xy,links_count,lwd=2){
  x<-xy$x;y<-xy$y
  border=0.25;M=1#max(links_count)
  plot(t(x),pch=1,xlim=c(min(c(x[1,],y[1,]))-border,max(c(x[1,],y[1,]))+border),
       ylim=c(min(c(x[2,],y[2,]))-border,max(c(x[2,],y[2,]))+border),xlab=" ",ylab=" ")
  for (i in 1:length(x[1,])){for (j in 1:length(y[1,])){
    if(links_count[i,j]/M>0.01){
      lines(x=c(x[1,i],y[1,j]),y=c(x[2,i],y[2,j]),
            col=gray(gray_scale(links_count[i,j]/M)),lwd=lwd) 
    }
  }}
  for (i in 1:length(x[1,])){for (j in 1:length(y[1,])){
    if (links_count[i,j]/M>0.2){
      lines(x=c(x[1,i],y[1,j]),y=c(x[2,i],y[2,j]),
            col=gray(gray_scale(links_count[i,j]/M)),lwd=lwd)
    }
  }}
  for (i in 1:length(x[1,])){
    points(x[1,i],x[2,i],pch=1,col='red')
  }
  for (j in 1:length(y[1,])){
    points(y[1,j],y[2,j],pch=4,col='blue')
  }
  legend.col(col = gray(gray_scale(c(0:100)/100)), lev =c(c(0:100)/100))
}
gray_scale=function(x){
  return (1-x)
}
legend.col <- function(col, lev){
  opar <- par
  n <- length(col)
  bx <- par("usr")
  box.cx <- c(bx[2] + (bx[2] - bx[1]) / 1000,
              bx[2] + (bx[2] - bx[1]) / 1000 + (bx[2] - bx[1]) / 50)
  box.cy <- c(bx[3], bx[3])
  box.sy <- (bx[4] - bx[3]) / n
  xx <- rep(box.cx, each = 2)
  par(xpd = TRUE)
  xx<-xx*0.98
  for(i in 1:n){
    yy <- c(box.cy[1] + (box.sy * (i - 1)),
            box.cy[1] + (box.sy * (i)),
            box.cy[1] + (box.sy * (i)),
            box.cy[1] + (box.sy * (i - 1)))
    polygon(xx, yy, col = col[i], border = col[i])
  }
  par(new = TRUE)
  plot(0, 0, type = "n",
       ylim = c(min(lev), max(lev)),
       yaxt = "n", ylab = "",
       xaxt = "n", xlab = "",
       frame.plot = FALSE)
  axis(side = 4, las = 2, tick = FALSE, line = -.5)
  par <- opar
}

###############################################
###########  FUNCTION FOR K COLORS ############
###############################################

#### Metropolis-within-Gibbs algorithm for ###############
#### the Poisson model with k colors (general case) ######
###
kd_MH_poisson_model<-function(points,n_points,T,burn_time,n_grid=200,border=10^5,starting_state=NULL,start_p_c=NULL,start_sigma=NULL,start_lambda=NULL,thinning=1,num_steps_in_2d=1,sample_sigma=TRUE,sample_p_c=TRUE,sample_lambda=TRUE,cut_density_flag=TRUE,plot_density=TRUE,eval_links_count=TRUE,sampler_2d=sampler_2d_P4,edge_correction=NULL,eval_association_count=TRUE,coast_distance=3000,wales_fraction_cutoff=1.9){
  # This function runs a Metropolis-within-Gibbs algorithm to sample from the Poisson model described in
  # Section 2 of the paper (Zanella 2014). The algorithm is described in Section 4.2 of (Zanella 2014).
  #INITIALIZES STATE AND PARAMETERS FOR THE MCMC
  k<-length(n_points)-1
  if (length(start_p_c)==0){p_c<-rdirichlet(n=1,par=alpha_p_c_prior(k))
  }else{p_c<-start_p_c}
  if (length(start_sigma)==0){sigma<-runif(1,min=0,max=sigma_max)
  }else{sigma<-start_sigma}
  if (length(start_lambda)==0){lambda<-rgamma(n=1,shape=k_lambda,scale=theta_lambda)
  }else{lambda<-start_lambda}
  if (length(starting_state)==0){#No starting state given: INITIALIZING with no links
    empty_matrix<-matrix(c(1:k),nrow=k,ncol=1)[,-1]
    starting_state<-list(color=rep(0,times=n_points[k+1]),cluster=empty_matrix)
  }
  state<-starting_state
  x_all<-NULL
  y_all<-NULL
  marks<-c()
  for (i in 1:k){
    x_all<-c(x_all,points[[i]][1,])
    y_all<-c(y_all,points[[i]][2,])
    marks<-c(marks,rep(names_included[i],length(points[[i]])/2))
  }
  xmin<-min(x_all)-border
  xmax<-max(x_all)+border
  ymin<-min(y_all)-border
  ymax<-max(y_all)+border
  ###### EVALUATES GAUSSIAN KERNEL SMOOTHING ESTIMATE OF THE CENTERS DENSITY g(z)
  points.dataframe<-data.frame(x=x_all,y=y_all,marks=marks)
  points.ppp<-as.ppp(points.dataframe,W=owin(xrange=c(xmin,xmax),yrange=c(ymin,ymax)))
  centre_density<-density(x=points.ppp,sigma=bw.scott(points.ppp),dimyx=c(n_grid,n_grid),edge=FALSE)#bw.relrisk(points.ppp,method="likelihood")
  if (cut_density_flag){
    centre_density<-cut_density(centre_density=centre_density,n_grid=n_grid,coast_distance=coast_distance,wales_fraction_cutoff=wales_fraction_cutoff)
    points.ppp<-as.ppp(points.dataframe,W=owin(xrange=c(xmin,xmax),yrange=c(ymin,ymax),mask=centre_density$v>1e-20))
    centre_density<-density(x=points.ppp,sigma=bw.relrisk(points.ppp,method="likelihood"),dimyx=c(n_grid,n_grid),diggle=TRUE)    
    centre_density$v[is.na(centre_density$v)]<-1e-100
  }      
  centre_density$v<-centre_density$v/(mean(centre_density$v)*(ymax-ymin)*(xmax-xmin))
  if(plot_density){
    image(centre_density,col=tim.colors(64))
    points(x=x_all,y=y_all)
  }
  centre_density<-list(x=centre_density$xcol,y=centre_density$yrow,z=t(centre_density$v),xmin=xmin,xmax=xmax,ymin=ymin,ymax=ymax,n_grid=n_grid)
  ###### INITIALIZES HISTORY VECTORS THAT WILL BE OUTPUT #####
  links_count<-kd_cluster_count_output(state=state,links_count=NULL,T=T,n_points=n_points)
  association_count<-matrix(rep(0,k^2),k,k)
  sigma_vec  <- rep(NA,(T+burn_time)%/%thinning+1)
  lambda_vec <- rep(NA,(T+burn_time)%/%thinning+1)
  p_c_vec    <- matrix(rep(NA , k*((T+burn_time)%/%thinning+1)) , nrow=k , ncol=(T+burn_time)%/%thinning+1)
  Y_vec    <- matrix(rep(0 , k*((T+burn_time)%/%thinning+1)) , nrow=k , ncol=(T+burn_time)%/%thinning+1)
  differences_from_ref<-rep(NA,(T+burn_time)%/%thinning+1)
  acceptance_rates<- rep(NA,(T+burn_time)%/%thinning)
  sigma_vec[1]  <- sigma
  lambda_vec[1] <- lambda
  p_c_vec[,1]   <- p_c
  if(length(state$cluster)>0){
    current_Y<-table(colSums(state$cluster!=0))
    Y_vec[as.double(names(current_Y)),1]<-current_Y*as.double(names(current_Y))
  }
  Y_vec[1,1]<-n_points[k+1]-sum(Y_vec[,1])
  differences_from_ref[1] <- sum(state$cluster!=0)
  c_vec<-c_vec_poisson(k=k,p_c=p_c,sigma=sigma,lambda=lambda)
  ###### STEP ZERO OF THE MCMC: MAKE MANY MOVES ON THE PARTITION WHILE KEEPING FIXED THE OTHER PARAMETERS
  for (iter in 1:10){
    colors_picked<-sort(sample.int(k,floor(k/2)))
    xy<-reduction_to_2d(colors_picked=colors_picked,state=state,n_points=n_points,points=points,x_all=x_all,y_all=y_all) #REDUCE TO 2-D the points
    weights<-build_weights_matrix_2dkd_poisson_inhom(xy=xy,centre_density=centre_density,c_vec=c_vec,sigma=sigma)
    state<-sampler_2d(state=state,weights=weights,xy=xy,num_steps_in_2d=num_steps_in_2d,colors_picked=colors_picked,n_points=n_points)
  }  
  ###### MCMC ALGORITHM
  for(t in 1:(T+burn_time)){
    #SELECTS THE SUBSET OF COLORS (DENOTED BY A IN THE PAPER)
    colors_picked<-sample.int(k,floor(k/2))
    #PROJECT ON THE 2COLOR CONFIGURATION AND EVALUATES THE WEIGHTS IN THE 2COLOR SPACE
    xy<-reduction_to_2d(colors_picked=colors_picked,state=state,n_points=n_points,points=points,x_all=x_all,y_all=y_all) #REDUCE TO 2-D the points
    weights<-build_weights_matrix_2dkd_poisson_inhom(xy=xy,centre_density=centre_density,c_vec=c_vec,sigma=sigma)
    #PERFORMS num_steps_in_2d STEPS IN THE 2COLOR SPACE
    state<-sampler_2d(state=state,weights=weights,xy=xy,num_steps_in_2d=num_steps_in_2d,colors_picked=colors_picked,n_points=n_points)
    #SAMPLES SIGMA, p^{(c)}, AND LAMBDA
    if (sample_sigma){
      n_minus_N<-(sum(state$cluster!=0)-dim(state$cluster)[2])
      sum_cl_sq_dist<-eval_sum_cl_sq_dist(cluster=state$cluster,n_points=n_points,points=points)
      for (iter_sigma in 1:10){
        sigma_new<-sigma+runif(1,min=-500,max=+500)
        if(sigma_new>0 & sigma_new<sigma_max){
          if(runif(1)<(sigma_new/sigma)^(-2*n_minus_N)*exp((-1/(sigma_new^2)+1/(sigma^2))*(sum_cl_sq_dist/2))){
            sigma<-sigma_new
          }
        }
      }
    }
    N_vec<-evaluate_N_vec(cluster=state$cluster,total_n_points=n_points[k+1])
    if (sample_p_c){
      p_c<-rdirichlet(n=1,par=alpha_p_c_prior(k)+N_vec)
    }
    if (sample_lambda){
      lambda<-rgamma(n=1,shape=posterior_shape_lambda(shape_lambda,N_vec),scale=posterior_scale_lambda(scale_lambda))
    }
    c_vec<-c_vec_poisson(k=k,p_c=p_c,sigma=sigma,lambda=lambda)
    ##### UPDATES HISTORY VECTORS THAT WILL BE OUTPUT
    if(t%%thinning==0){
      sigma_vec[t%/%thinning+1]  <- sigma
      lambda_vec[t%/%thinning+1] <- lambda
      p_c_vec[,t%/%thinning+1]   <- p_c
      if(length(state$cluster)>0){
        current_Y<-table(colSums(state$cluster!=0))
        Y_vec[as.double(names(current_Y)),t%/%thinning+1]<-current_Y*as.double(names(current_Y))
      }
      Y_vec[1,t%/%thinning+1]<-n_points[k+1]-sum(Y_vec[,t%/%thinning+1])
      differences_from_ref[t%/%thinning+1]<-sum(state$cluster!=0)
      if(t>burn_time & eval_links_count){
        links_count<-kd_cluster_count_output(state=state,links_count=links_count,T=T%/%thinning,n_points=n_points)
      }
      if(t>burn_time & eval_association_count){
        for (A in 2:k){
          for (B in 1:(A-1)){
            association_count[A,B]<-association_count[A,B]+
              sum(state$cluster[A,]>0&state$cluster[B,]>0)/(T%/%thinning*sum(N_vec))            
          }
        }
      }
    }
    if(t%%100==0){
      print(paste('time=',t))
    }
  }
  return(list(sigma_vec=sigma_vec,lambda_vec=lambda_vec,p_c_vec=p_c_vec,Y_vec=Y_vec,
              links_count=links_count,association_count=association_count,state=state,differences_from_ref=differences_from_ref,acceptance_rates=acceptance_rates))
}
kd_MH_poisson_model_sel_region<-function(points,n_points,T,burn_time,n_grid=200,border=10^5,starting_state=NULL,start_p_c=NULL,start_sigma=NULL,start_lambda=NULL,thinning=1,num_steps_in_2d=1,sample_sigma=TRUE,sample_p_c=TRUE,sample_lambda=TRUE,cut_density_flag=TRUE,plot_density=TRUE,eval_links_count=TRUE,sampler_2d=sampler_2d_P4,edge_correction=FALSE,select_region=FALSE,cutoff=NULL){
  # This function runs a Metropolis-within-Gibbs algorithm to sample from the Poisson model described in
  # Section 2 of the paper (Zanella 2014). The algorithm is described in Section 4.2 of (Zanella 2014).
  ###### EVALUATES GAUSSIAN KERNEL SMOOTHING ESTIMATE OF THE CENTERS DENSITY g(z)
  x_all<-NULL
  y_all<-NULL
  marks_list<-as.character(1:k)
  marks<-c()
  for (i in 1:k){
    x_all<-c(x_all,points[[i]][1,])
    y_all<-c(y_all,points[[i]][2,])
    marks<-c(marks,rep(marks_list[i],length(points[[i]])/2))
  }
  xmin<-min(x_all)-border
  xmax<-max(x_all)+border
  ymin<-min(y_all)-border
  ymax<-max(y_all)+border
  points.dataframe<-data.frame(x=x_all,y=y_all,marks=marks)
  points.ppp<-as.ppp(points.dataframe,W=owin(xrange=c(xmin,xmax),yrange=c(ymin,ymax)))
  #centre_density<-density(x=points.ppp,sigma=bw.relrisk(points.ppp,method='likelihood'),dimyx=c(n_grid,n_grid),edge=FALSE)
  centre_density<-kde2d(x=x_all,y=y_all,n=n_grid,lims=c(xmin,xmax,ymin,ymax),h=16*10^04)
  #   if (cut_density_flag){
  #     centre_density<-cut_density(centre_density=centre_density,n_grid=n_grid)
  #   }
  #   if (edge_correction){
  #     points.ppp<-as.ppp(points.dataframe,W=owin(xrange=c(xmin,xmax),yrange=c(ymin,ymax),mask=centre_density$v>1e-20))
  #     centre_density<-density(x=points.ppp,sigma=bw.relrisk(points.ppp,method='likelihood'),dimyx=c(n_grid,n_grid),diggle=TRUE)    
  #     centre_density$v[is.na(centre_density$v)]<-1e-100
  #   }
  if (select_region){
    selected_region_small<-owin(xrange=c(xmin,xmax),yrange=c(ymin,ymax),mask=t(centre_density$z)>max(centre_density$z)*cutoff)
    selected_region_big<-owin(xrange=c(xmin,xmax),yrange=c(ymin,ymax),mask=t(centre_density$z)>max(centre_density$z)*0.98*cutoff)#0.9
    centre_density$z[t(!selected_region_big$m)]<-0
    #     bw.used<-bw.relrisk(points.ppp,method='likelihood')
    #     selected_region_small<-owin(xrange=c(xmin,xmax),yrange=c(ymin,ymax),mask=centre_density$v>max(centre_density$v)*cutoff)
    #     selected_region_big<-owin(xrange=c(xmin,xmax),yrange=c(ymin,ymax),mask=centre_density$v>max(centre_density$v)*0.98*cutoff)
    #     centre_density<-density(x=points.ppp[selected_region_big],sigma=bw.used,dimyx=c(n_grid,n_grid),diggle=TRUE)    
    #     centre_density$v[is.na(centre_density$v)]<-1e-100
    points<-list()
    points[[k]]<-NA
    points.ppp<-points.ppp[selected_region_small]
    for (i in 1:k){
      selected.ppp<-as.ppp(X=points.ppp[(points.ppp$marks == marks_list[i])])
      points[[i]]<-rbind(selected.ppp$x,selected.ppp$y)
    }
    n_points<-rep(0,k+1)
    for (i in 1:k){
      n_points[(i+1):(k+1)]<-n_points[(i+1):(k+1)]+length(points[[i]])/2
    }
    x_all<-NULL
    y_all<-NULL
    marks_list<-as.character(1:k)
    marks<-c()
    for (i in 1:k){
      x_all<-c(x_all,points[[i]][1,])
      y_all<-c(y_all,points[[i]][2,])
      marks<-c(marks,rep(marks_list[i],length(points[[i]])/2))
    }
  }
  #   centre_density$v<-centre_density$v/(mean(centre_density$v)*(ymax-ymin)*(xmax-xmin))
  centre_density$z<-centre_density$z/(mean(centre_density$z)*(ymax-ymin)*(xmax-xmin))
  if(plot_density){
    image(centre_density,col=tim.colors(64))
    points(x=x_all,y=y_all)
    print(n_points)
  }
  #   centre_density<-list(x=centre_density$xcol,y=centre_density$yrow,z=t(centre_density$v),xmin=xmin,xmax=xmax,ymin=ymin,ymax=ymax,n_grid=n_grid)
  centre_density<-list(x=centre_density$x,y=centre_density$y,z=centre_density$z,xmin=xmin,xmax=xmax,ymin=ymin,ymax=ymax,n_grid=n_grid)
  print(range(centre_density$z))
  print(mean(centre_density$z))
  #INITIALIZES STATE AND PARAMETERS FOR THE MCMC
  k<-length(n_points)-1
  if (length(start_p_c)==0){p_c<-rdirichlet(n=1,par=alpha_p_c_prior(k))
  }else{p_c<-start_p_c}
  if (length(start_sigma)==0){sigma<-sqrt(rigamma(n=1,a=shape_sigma,b=scale_sigma))
  }else{sigma<-start_sigma}
  if (length(start_lambda)==0){lambda<-rgamma(n=1,shape=shape_lambda,scale=scale_lambda)
  }else{lambda<-start_lambda}
  if (length(starting_state)==0){#No starting state given: INITIALIZING with no links
    empty_matrix<-matrix(c(1:k),nrow=k,ncol=1)[,-1]
    starting_state<-list(color=rep(0,times=n_points[k+1]),cluster=empty_matrix)
  }
  state<-starting_state
  c_vec<-c_vec_poisson(k=k,p_c=p_c,sigma=sigma,lambda=lambda)
  ###### INITIALIZES HISTORY VECTORS THAT WILL BE OUTPUT #####
  links_count<-kd_cluster_count_output(state=state,links_count=NULL,T=T,n_points=n_points)
  sigma_vec  <- rep(NA,(T+burn_time)%/%thinning+1)
  lambda_vec <- rep(NA,(T+burn_time)%/%thinning+1)
  p_c_vec    <- matrix(rep(NA , k*((T+burn_time)%/%thinning+1)) , nrow=k , ncol=(T+burn_time)%/%thinning+1)
  Y_vec    <- matrix(rep(0 , k*((T+burn_time)%/%thinning+1)) , nrow=k , ncol=(T+burn_time)%/%thinning+1)
  differences_from_ref<-rep(NA,(T+burn_time)%/%thinning+1)
  sigma_vec[1]  <- sigma
  lambda_vec[1] <- lambda
  p_c_vec[,1]   <- p_c
  if(length(state$cluster)>0){
    current_Y<-table(colSums(state$cluster!=0))
    Y_vec[as.double(names(current_Y)),1]<-current_Y*as.double(names(current_Y))
  }
  Y_vec[1,1]<-n_points[k+1]-sum(Y_vec[,1])
  differences_from_ref[1] <- sum(state$cluster!=0)
  ###### STEP ZERO OF THE MCMC: MAKE MANY MOVES ON THE PARTITION WHILE KEEPING FIXED THE OTHER PARAMETERS
  for (iter in 1:50){
    colors_picked<-sort(sample.int(k,floor(k/2)))
    xy<-reduction_to_2d(colors_picked=colors_picked,state=state,n_points=n_points,points=points,x_all=x_all,y_all=y_all) #REDUCE TO 2-D the points
    weights<-build_weights_matrix_2dkd_poisson_inhom(xy=xy,centre_density=centre_density,c_vec=c_vec,sigma=sigma)
    state<-sampler_2d(state=state,weights=weights,xy=xy,num_steps_in_2d=num_steps_in_2d,colors_picked=colors_picked,n_points=n_points)
  }
  ###### MCMC ALGORITHM
  for(t in 1:(T+burn_time)){
    #SELECTS THE SUBSET OF COLORS (DENOTED BY A IN THE PAPER)
    colors_picked<-sample.int(k,floor(k/2))
    #PROJECT ON THE 2COLOR CONFIGURATION AND EVALUATES THE WEIGHTS IN THE 2COLOR SPACE
    xy<-reduction_to_2d(colors_picked=colors_picked,state=state,n_points=n_points,points=points,x_all=x_all,y_all=y_all)
    weights<-build_weights_matrix_2dkd_poisson_inhom(xy=xy,centre_density=centre_density,c_vec=c_vec,sigma=sigma)
    #PERFORMS num_steps_in_2d STEPS IN THE 2COLOR SPACE
    state<-sampler_2d(state=state,weights=weights,xy=xy,num_steps_in_2d=num_steps_in_2d,colors_picked=colors_picked,n_points=n_points)
    #SAMPLES SIGMA, p^{(c)}, AND lambda
    if (sample_sigma){
      n_minus_N<-(sum(state$cluster!=0)-dim(state$cluster)[2])
      sum_cl_sq_dist<-eval_sum_cl_sq_dist(cluster=state$cluster,n_points=n_points,points=points)
      for (iter_sigma in 1:10){
        sigma_new<-sigma+runif(1,min=-500,max=+500)
        if(sigma_new>0 & sigma_new<sigma_max){
          if(runif(1)<(sigma_new/sigma)^(-2*n_minus_N)*exp((-1/(sigma_new^2)+1/(sigma^2))*(sum_cl_sq_dist/2))){
            sigma<-sigma_new
          }
        }
      }
    }
    N_vec<-evaluate_N_vec(cluster=state$cluster,total_n_points=n_points[k+1])
    if (sample_p_c){
      p_c<-rdirichlet(n=1,par=posterior_alpha_p_c(k,N_vec))
    }
    if (sample_lambda){
      lambda<-rgamma(n=1,shape=posterior_shape_lambda(shape_lambda,N_vec),scale=posterior_scale_lambda(scale_lambda))
    }
    c_vec<-c_vec_poisson(k=k,p_c=p_c,sigma=sigma,lambda=lambda)
    ##### UPDATES HISTORY VECTORS THAT WILL BE OUTPUT
    if(t%%thinning==0){
      sigma_vec[t%/%thinning+1]  <- sigma
      lambda_vec[t%/%thinning+1] <- lambda
      p_c_vec[,t%/%thinning+1]   <- p_c
      if(length(state$cluster)>0){
        current_Y<-table(colSums(state$cluster!=0))
        Y_vec[as.double(names(current_Y)),t%/%thinning+1]<-current_Y*as.double(names(current_Y))
      }
      Y_vec[1,t%/%thinning+1]<-n_points[k+1]-sum(Y_vec[,t%/%thinning+1])
      differences_from_ref[t%/%thinning+1]<-sum(state$cluster!=0)
      if(t>burn_time & eval_links_count){
        links_count<-kd_cluster_count_output(state=state,links_count=links_count,T=T%/%thinning,n_points=n_points)
      }
    }
    if(t%%100==0){
      print(paste('time=',t))
    }
  }
  return(list(sigma_vec=sigma_vec,lambda_vec=lambda_vec,p_c_vec=p_c_vec,Y_vec=Y_vec,
              links_count=links_count,state=state,differences_from_ref=differences_from_ref))
}



#### AUXILIARY FUNCTIONS FOR THE METROPOLIS STEP OF THE METROPOLIS-WITHIN-GIBBS ALGORITHM #######
###
reduction_to_2d<-function(colors_picked,state,n_points,points,x_all,y_all){
  k<-length(n_points)-1
  s1<-length(colors_picked)
  remained_colors<-c(1:k)[-colors_picked]
  s2<-length(remained_colors)
  x<-c()
  y<-c()
  x_marks<-c()
  y_marks<-c()
  kd_indices_x<-c()
  kd_indices_y<-c()
  twod_matching<-c()
  if(length(state$cluster)>0){
    x_marks_long<-colSums(matrix(state$cluster[colors_picked,],nrow=s1)!=0)
    y_marks_long<-colSums(matrix(state$cluster[remained_colors,],nrow=s2)!=0)
    cluster_inf<-state$cluster
    cluster_inf[cluster_inf==0]<-Inf
    first_cord_mat<-state$cluster
    first_cord_mat[first_cord_mat>0]<-x_all[first_cord_mat]
    second_cord_mat<-state$cluster
    second_cord_mat[second_cord_mat>0]<-y_all[second_cord_mat]
    if(any(x_marks_long>0)){
      x_marks<-x_marks_long[x_marks_long>0]
      x<-rbind(colSums(matrix(first_cord_mat[colors_picked,x_marks_long>0],nrow=s1))/x_marks,
               colSums(matrix(second_cord_mat[colors_picked,x_marks_long>0],nrow=s1))/x_marks)
      mat<-matrix(cluster_inf[colors_picked,x_marks_long>0],nrow=length(colors_picked))
      kd_indices_x<-do.call(pmin, lapply(1:nrow(mat), function(i)mat[i,]))# apply(matrix(state$cluster[colors_picked,x_marks_long>0],nrow=s1),2,function(vec){vec[vec!=0][1]})
    }
    if(any(y_marks_long>0)){
      y_marks<-y_marks_long[y_marks_long>0]
      y<-rbind(colSums(matrix(first_cord_mat[remained_colors,y_marks_long>0],nrow=s2))/y_marks,
               colSums(matrix(second_cord_mat[remained_colors,y_marks_long>0],nrow=s2))/y_marks)
      mat<-matrix(cluster_inf[remained_colors,y_marks_long>0],nrow=length(remained_colors))
      kd_indices_y<-do.call(pmin, lapply(1:nrow(mat), function(i)mat[i,]))#apply(matrix(state$cluster[remained_colors,y_marks_long>0],nrow=s2),2,function(vec){vec[vec!=0][1]})
    }
    if(any(x_marks_long>0)&any(y_marks_long>0)){
      twod_matching<-rep(0,length(x_marks_long>0))
      twod_matching[x_marks_long>0&y_marks_long>0]<-which(x_marks_long[y_marks_long>0]>0)
      twod_matching<-twod_matching[x_marks_long>0]
    }
  }
  #THEN PUT FREE POINTS
  for (i in colors_picked){
    x<-cbind(x,points[[i]][,which(state$color[(n_points[i]+1):(n_points[i+1])]==0)])
    kd_indices_x<-c(kd_indices_x,n_points[i]+which(state$color[(n_points[i]+1):(n_points[i+1])]==0))
  }
  x_marks<-c(x_marks,rep(1,length(x[1,])-length(x_marks)))
  twod_matching<-c(twod_matching,rep(0,length(x[1,])-length(twod_matching)))
  for (i in remained_colors){
    y<-cbind(y,points[[i]][,which(state$color[(n_points[i]+1):(n_points[i+1])]==0)])
    kd_indices_y<-c(kd_indices_y,n_points[i]+which(state$color[(n_points[i]+1):(n_points[i+1])]==0))
  }
  y_marks<-c(y_marks,rep(1,length(y[1,])-length(y_marks)))
  deltas2<-(rep(y[1,],each=length(x[1,]))-rep(x[1,],length(y[1,])))^2+(rep(y[2,],each=length(x[2,]))-rep(x[2,],length(y[2,])))^2
  x_marks_matrix<-rep(x_marks,length(y[1,]))#matrix(rep(x_marks,length(y[1,])),nrow=length(x[1,]),ncol=length(y[1,]))
  y_marks_matrix<-rep(y_marks,each=length(x[1,]))#matrix(rep(y_marks,length(x[1,])),nrow=length(x[1,]),ncol=length(y[1,]),byrow=TRUE)
  deltas2_corr<-(deltas2/2)* x_marks_matrix*y_marks_matrix/(x_marks_matrix+y_marks_matrix)
  xy<-list(x=x,y=y,x_marks=x_marks,y_marks=y_marks,kd_indices_x=kd_indices_x,
           kd_indices_y=kd_indices_y,twod_matching=twod_matching,deltas2_corr=deltas2_corr)
  return(xy)
}
sampler_2d_P3<-function(state,weights,xy,num_steps_in_2d,colors_picked,n_points){  
  matching<-xy$twod_matching
  prop_mat<-q_function(build_post_mat(weights=weights,matching=matching))
  norm_const<-sum(prop_mat)
  marginal_i<-rowSums(prop_mat)
  for(t in 1:num_steps_in_2d){
    #MAKING THE PROPOSAL AND EVALUATING THE ACCEPTANCE PROBABILITY
    cum_probs<-cumsum(marginal_i)
    i<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    cum_probs<-cumsum(prop_mat[i,])
    j<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    new_matching<-matching_proposed(matching=matching,i=i,j=j)
    new_prop_row_i<-q_function(update_row(weights=weights,i=i,current_matching=new_matching))
    new_prop_col_j<-q_function(update_column(weights=weights,j=j,current_matching=new_matching))
    if (matching[i]==j | (matching[i]==0 & sum(matching==j)==0)){#ADDITION OR DELETION MOVE
      delta_i<-sum(new_prop_row_i[-j])-sum(prop_mat[i,-j])
      new_norm_const<-norm_const+delta_i+sum(new_prop_col_j)-sum(prop_mat[,j])
      ap<-norm_const/new_norm_const
      if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
        marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]
        marginal_i[i]<-marginal_i[i]+delta_i
        prop_mat[i,]<-new_prop_row_i
        prop_mat[,j]<-new_prop_col_j
        norm_const<-new_norm_const
        state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
        matching<-new_matching
      }
    }else{
      if (matching[i]==0 & sum(matching==j)==1){#SWITCH MOVE ('PIVOT' ON y)
        i_prime<-which(matching==j)
        new_prop_row_i_prime<-q_function(update_row(weights=weights,i=i_prime,current_matching=new_matching))
        delta_i<-sum(new_prop_row_i[-j])-sum(prop_mat[i,-j])
        delta_i_prime<-sum(new_prop_row_i_prime[-j])-sum(prop_mat[i_prime,-j])
        new_norm_const<-norm_const+delta_i+delta_i_prime+sum(new_prop_col_j)-sum(prop_mat[,j])
        ap<-norm_const/new_norm_const
        if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
          marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]
          marginal_i[i]<-marginal_i[i]+delta_i
          marginal_i[i_prime]<-marginal_i[i_prime]+delta_i_prime
          prop_mat[i,]<-new_prop_row_i
          prop_mat[i_prime,]<-new_prop_row_i_prime
          prop_mat[,j]<-new_prop_col_j
          norm_const<-new_norm_const
          state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
          matching<-new_matching
        }
      }else{
        if (matching[i]!=0 & sum(matching==j)==0){#SWITCH MOVE ('PIVOT' ON X)
          delta_i<-sum(new_prop_row_i[-c(j,matching[i])])-sum(prop_mat[i,-c(j,matching[i])])
          new_prop_col_j_prime<-q_function(update_column(weights=weights,j=matching[i],current_matching=new_matching))
          new_norm_const<-norm_const+delta_i+sum(new_prop_col_j)-sum(prop_mat[,j])+sum(new_prop_col_j_prime)-sum(prop_mat[,matching[i]])
          ap<-norm_const/new_norm_const
          if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
            marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]+new_prop_col_j_prime-prop_mat[,matching[i]]
            marginal_i[i]<-marginal_i[i]+delta_i
            prop_mat[i,]<-new_prop_row_i
            prop_mat[,j]<-new_prop_col_j
            prop_mat[,matching[i]]<-new_prop_col_j_prime
            norm_const<-new_norm_const
            state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
            matching<-new_matching
          }
        }else{
          if (matching[i]!=0 & sum(matching==j)==1){#DOUBLE-SWITCH MOVE
            delta_i<-sum(new_prop_row_i[-c(j,matching[i])])-sum(prop_mat[i,-c(j,matching[i])])
            i_prime<-which(matching==j)
            new_prop_row_i_prime<-q_function(update_row(weights=weights,i=i_prime,current_matching=new_matching))
            delta_i_prime<-sum(new_prop_row_i_prime[-c(j,matching[i])])-sum(prop_mat[i_prime,-c(j,matching[i])])
            new_prop_col_j_prime<-q_function(update_column(weights=weights,j=matching[i],current_matching=new_matching))
            new_norm_const<-norm_const+delta_i+delta_i_prime+sum(new_prop_col_j)-sum(prop_mat[,j])+sum(new_prop_col_j_prime)-sum(prop_mat[,matching[i]])
            ap<-norm_const/new_norm_const
            if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
              marginal_i<-marginal_i+new_prop_col_j-prop_mat[,j]+new_prop_col_j_prime-prop_mat[,matching[i]]
              marginal_i[i]<-marginal_i[i]+delta_i
              marginal_i[i_prime]<-marginal_i[i_prime]+delta_i_prime
              prop_mat[i,]<-new_prop_row_i
              prop_mat[i_prime,]<-new_prop_row_i_prime
              prop_mat[,j]<-new_prop_col_j
              prop_mat[,matching[i]]<-new_prop_col_j_prime
              norm_const<-new_norm_const
              state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
              matching<-new_matching
            }  
          }
        }
      }
    }
  }
  return(state)
}
sampler_2d_P4<-function(state,weights,xy,num_steps_in_2d,colors_picked,n_points){
  #INITIALIZATION
  matching<-xy$twod_matching
  q_add_mat<-build_q_add_mat(weights)
  q_del_mat<-build_q_rem_mat(weights)
  prop_mat<-q_add_mat
  links_indices<-matrix(c(which(matching>0),matching[matching>0]),ncol=2)
  prop_mat[links_indices]<-q_del_mat[links_indices]
  marginal_i<-rowSums(prop_mat)
  norm_const<-sum(marginal_i)
  for(t in 1:num_steps_in_2d){
    cum_probs<-cumsum(marginal_i)
    i<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    cum_probs<-cumsum(prop_mat[i,])
    j<-which(cum_probs>(runif(1)*cum_probs[length(cum_probs)]))[1]
    if (matching[i]==j){#CASE 1: x&y ARE MATCHED TOGHETER: REMOVE THE LINK
      post_ratio<-1/weights[i,j]
      delta_norm_const<-q_add_mat[i,j]-q_del_mat[i,j]
      ap<-post_ratio*q_add_mat[i,j]/(norm_const+delta_norm_const)/(q_del_mat[i,j]/norm_const)
      if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
        prop_mat[i,j]<-q_add_mat[i,j]
        marginal_i[i]<-marginal_i[i]+delta_norm_const
        norm_const<-norm_const+delta_norm_const
        state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
        matching<-matching_proposed(matching=matching,i=i,j=j)
      }
    }else{
      if (matching[i]==0 & sum(matching==j)==0){#CASE 2: x&y ARE FREE: ADD THE LINK
        post_ratio<-weights[i,j]
        delta_norm_const<-q_del_mat[i,j]-q_add_mat[i,j]
        ap<-post_ratio*q_del_mat[i,j]/(norm_const+delta_norm_const)/(q_add_mat[i,j]/norm_const)
        if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
          prop_mat[i,j]<-q_del_mat[i,j]
          marginal_i[i]<-marginal_i[i]+delta_norm_const
          norm_const<-norm_const+delta_norm_const
          state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
          matching<-matching_proposed(matching=matching,i=i,j=j)
        }        
      }else{
        if (matching[i]==0 & sum(matching==j)==1){#CASE 3: x IS FREE y IS BUSY: 'PIVOT' ON y
          i_prime<-which(matching==j)
          post_ratio<-weights[i,j]/weights[i_prime,j]
          delta_norm_const1<-q_del_mat[i,j]-q_add_mat[i,j]  
          delta_norm_const2<-q_add_mat[i_prime,j]-q_del_mat[i_prime,j]
          ap<-post_ratio*q_add_mat[i_prime,j]/(norm_const+delta_norm_const1+delta_norm_const2)/(q_add_mat[i,j]/norm_const)
          if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
            prop_mat[i,j]<-q_del_mat[i,j]
            prop_mat[i_prime,j]<-q_add_mat[i_prime,j]
            marginal_i[i]<-marginal_i[i]+delta_norm_const1   
            marginal_i[i_prime]<-marginal_i[i_prime]+delta_norm_const2
            norm_const<-norm_const+delta_norm_const1+delta_norm_const2
            state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
            matching<-matching_proposed(matching=matching,i=i,j=j)
          }
        }else{
          if (matching[i]!=0 & sum(matching==j)==0){#CASE 4: x IS BUSY y IS FREE: 'PIVOT' ON x
            post_ratio<-weights[i,j]/weights[i,matching[i]]
            delta_norm_const<-q_del_mat[i,j]-q_add_mat[i,j]+q_add_mat[i,matching[i]]-q_del_mat[i,matching[i]] 
            ap<-post_ratio*q_add_mat[i,matching[i]]/(norm_const+delta_norm_const)/(q_add_mat[i,j]/norm_const)
            if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
              prop_mat[i,j]<-q_del_mat[i,j]
              prop_mat[i,matching[i]]<-q_add_mat[i,matching[i]]
              marginal_i[i]<-marginal_i[i]+delta_norm_const
              norm_const<-norm_const+delta_norm_const
              state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
              matching<-matching_proposed(matching=matching,i=i,j=j)
            }
          }else{
            if (matching[i]!=0 & sum(matching==j)==1){#CASE 5: x&y ARE BOTH BUSY: SWITCH THE LINKS 
              i_prime<-which(matching==j)
              post_ratio<-weights[i,j]*weights[i_prime,matching[i]]/
                (weights[i,matching[i]]*weights[i_prime,j])
              delta_norm_const1<-q_del_mat[i,j]-q_add_mat[i,j]+q_add_mat[i,matching[i]]-q_del_mat[i,matching[i]]
              delta_norm_const2<-q_del_mat[i_prime,matching[i]]-q_add_mat[i_prime,matching[i]]+q_add_mat[i_prime,j]-q_del_mat[i_prime,j]
              ap<-post_ratio*(q_add_mat[i_prime,j]+q_add_mat[i,matching[i]])/(norm_const+delta_norm_const1+delta_norm_const2)/((q_add_mat[i,j]+q_add_mat[i_prime,matching[i]])/norm_const)
              if (runif(1)<ap){#ACCEPT THE MOVE WITH ACCEPTING PROBABILITY
                prop_mat[i,j]<-q_del_mat[i,j]
                prop_mat[i,matching[i]]<-q_add_mat[i,matching[i]]
                prop_mat[i_prime,matching[i]]<-q_del_mat[i_prime,matching[i]]
                prop_mat[i_prime,j]<-q_add_mat[i_prime,j]
                marginal_i[i]<-marginal_i[i]+delta_norm_const1
                marginal_i[i_prime]<-marginal_i[i_prime]+delta_norm_const2                
                norm_const<-norm_const+delta_norm_const1+delta_norm_const2
                state<-update_state(state=state,twod_matching=matching,xy=xy,i=i,j=j,colors_picked=colors_picked,n_points=n_points)
                matching<-matching_proposed(matching=matching,i=i,j=j)
              }  
            }
          }
        }
      }
    }
  }
  return(state)
}
update_state<-function(state,twod_state=NULL,twod_matching=NULL,i,j,xy,colors_picked,n_points){
  if (length(twod_matching)==0){
    twod_matching<-twod_state$twod_matching
  }
  #INITIALIZATION
  i_kd<-xy$kd_indices_x[i]
  j_kd<-xy$kd_indices_y[j]
  cluster_i<-state$color[i_kd]
  cluster_j<-state$color[j_kd]
  color_i<-sum(n_points<i_kd)
  color_j<-sum(n_points<j_kd)
  k<-length(n_points)-1
  remaining_colors<-c(1:k)[-colors_picked]
  if (twod_matching[i]==j){#CASE 1: x&y ARE MATCHED TOGHETER: REMOVE THE LINK
    if(xy$x_marks[i]==1 & xy$y_marks[j]==1){#both clusters become single points: remove them from the cluster list
      state$color[c(i_kd,j_kd)]<-0
      state$cluster<-matrix(state$cluster[,-cluster_i],nrow=k,ncol=length(state$cluster[1,])-1)
      state$color[state$color>cluster_i]<-state$color[state$color>cluster_i]-1
    }else{if(xy$x_marks[i]==1){#x become a single point: remove them from the cluster list
      state$color[i_kd]<-0
      state$cluster[colors_picked,cluster_i]<-0
    }else{if(xy$y_marks[j]==1){#y become a single point: remove them from the cluster list
      state$color[j_kd]<-0
      state$cluster[remaining_colors,cluster_j]<-0
    }else{#none of them becomes a single point: split the cluster
      state$cluster<-cbind(state$cluster,state$cluster[,cluster_j])
      state$cluster[remaining_colors,cluster_i]<-0
      state$cluster[colors_picked,length(state$cluster[1,])]<-0
      state$color[  c(  state$cluster[remaining_colors, length(state$cluster[1,]) ]  )  ]<-length(state$cluster[1,])
    }
    }
    }
  }else{
    if (twod_matching[i]==0 & sum(twod_matching==j)==0){#CASE 2: x&y ARE FREE: ADD THE LINK
      if(xy$x_marks[i]==1 & xy$y_marks[j]==1){#both x and y were single points: add them to the cluster list
        state$color[c(i_kd,j_kd)]<-length(state$cluster[1,])+1
        state$cluster<-cbind(state$cluster,rep(0,k))
        state$cluster[color_i,length(state$cluster[1,])]<-i_kd
        state$cluster[color_j,length(state$cluster[1,])]<-j_kd  
      }else{if(xy$x_marks[i]==1){#x was a single point: add it to the cluster list
        state$color[i_kd]<-cluster_j
        state$cluster[color_i,cluster_j]<-i_kd
      }else{if(xy$y_marks[j]==1){#y was a single point: add it to the cluster list
        state$color[j_kd]<-cluster_i
        state$cluster[color_j,cluster_i]<-j_kd
      }else{#none of x and y was a single point: merge the two clusters
        state$cluster[remaining_colors,cluster_i]<-state$cluster[remaining_colors,cluster_j]
        state$color[c(state$cluster[remaining_colors,cluster_i])]<-cluster_i
        state$cluster<-matrix(state$cluster[,-cluster_j],nrow=k,ncol=length(state$cluster[1,])-1)
        state$color[state$color>cluster_j]<-state$color[state$color>cluster_j]-1
      }
      }
      }      
    }else{
      if (twod_matching[i]==0 & sum(twod_matching==j)==1){#CASE 3: x IS FREE y IS BUSY: 'PIVOT' ON y
        i_old<-which(twod_matching==j)
        i_kd_old<-xy$kd_indices_x[i_old]
        color_i_old<-sum(n_points<i_kd_old)
        if(xy$x_marks[i]==1 & xy$x_marks[i_old]==1){#both x and x_old were single points
          state$color[i_kd_old]<-0
          state$color[i_kd]<-cluster_j
          state$cluster[color_i_old,cluster_j]<-0
          state$cluster[color_i,cluster_j]<-i_kd
        }else{if(xy$x_marks[i]==1){#x was a single point: add it to the cluster list
          state$cluster<-cbind(state$cluster,rep(0,k))
          state$cluster[colors_picked,length(state$cluster[1,])]<-state$cluster[colors_picked,cluster_j]
          state$cluster[colors_picked,cluster_j]<-0
          state$cluster[color_i,cluster_j]<-i_kd
          state$color[i_kd]<-cluster_j
          state$color[state$cluster[colors_picked,length(state$cluster[1,])]]<-length(state$cluster[1,])
        }else{if(xy$x_marks[i_old]==1){#x_old becomes a single point: remove it from the cluster matrix
          state$color[i_kd_old]<-0  
          state$cluster[colors_picked,cluster_j]<-state$cluster[colors_picked,cluster_i]
          state$color[state$cluster[colors_picked,cluster_j]]<-cluster_j
          state$cluster<-matrix(state$cluster[,-cluster_i],nrow=k,ncol=length(state$cluster[1,])-1)
          state$color[state$color>cluster_i]<-state$color[state$color>cluster_i]-1
        }else{#none of x and x_old was a single point: switch the two clusters
          temp_vec<-state$cluster[colors_picked,cluster_i]
          state$cluster[colors_picked,cluster_i]<-state$cluster[colors_picked,cluster_j]
          state$color[c(state$cluster[colors_picked,cluster_i])]<-cluster_i
          state$cluster[colors_picked,cluster_j]<-temp_vec
          state$color[c(state$cluster[colors_picked,cluster_j])]<-cluster_j
        }
        }
        }      
      }else{
        if (twod_matching[i]!=0 & sum(twod_matching==j)==0){#CASE 4: x IS BUSY y IS FREE: 'PIVOT' ON x
          j_old<-twod_matching[i]
          j_kd_old<-xy$kd_indices_y[j_old]
          color_j_old<-sum(n_points<j_kd_old)
          if(xy$y_marks[j]==1 & xy$y_marks[j_old]==1){#both y and y_old were single points
            state$color[j_kd_old]<-0
            state$color[j_kd]<-cluster_i
            state$cluster[color_j_old,cluster_i]<-0
            state$cluster[color_j,cluster_i]<-j_kd
          }else{if(xy$y_marks[j]==1){#y was a single point, y_old was not
            state$cluster<-cbind(state$cluster,rep(0,k))
            state$cluster[remaining_colors,length(state$cluster[1,])]<-state$cluster[remaining_colors,cluster_i]
            state$cluster[remaining_colors,cluster_i]<-0
            state$cluster[color_j,cluster_i]<-j_kd
            state$color[j_kd]<-cluster_i
            state$color[state$cluster[remaining_colors,length(state$cluster[1,])]]<-length(state$cluster[1,])
          }else{if(xy$y_marks[j_old]==1){#y_old becomes a single point, y was not a single point
            state$color[j_kd_old]<-0  
            state$cluster[remaining_colors,cluster_i]<-state$cluster[remaining_colors,cluster_j]
            state$color[state$cluster[remaining_colors,cluster_i]]<-cluster_i
            state$cluster<-matrix(state$cluster[,-cluster_j],nrow=k,ncol=length(state$cluster[1,])-1)
            state$color[state$color>cluster_j]<-state$color[state$color>cluster_j]-1
          }else{#none of y and y_old was a single point: switch the two clusters
            temp_vec<-state$cluster[remaining_colors,cluster_j]
            state$cluster[remaining_colors,cluster_j]<-state$cluster[remaining_colors,cluster_i]
            state$color[c(state$cluster[remaining_colors,cluster_j])]<-cluster_j
            state$cluster[remaining_colors,cluster_i]<-temp_vec
            state$color[c(state$cluster[remaining_colors,cluster_i])]<-cluster_i
          }
          }
          }
        }else{
          if (twod_matching[i]!=0 & sum(twod_matching==j)==1){#CASE 5: x&y ARE BOTH BUSY: SWITCH THE LINKS 
            temp_vec<-state$cluster[remaining_colors,cluster_j]
            state$cluster[remaining_colors,cluster_j]<-state$cluster[remaining_colors,cluster_i]
            state$color[c(state$cluster[remaining_colors,cluster_j])]<-cluster_j
            state$cluster[remaining_colors,cluster_i]<-temp_vec
            state$color[c(state$cluster[remaining_colors,cluster_i])]<-cluster_i
          }
        }
      }
    }
  }
  return(state)
} 
build_weights_matrix_2dkd_poisson_inhom<-function(xy,centre_density,c_vec,sigma){
  x<-xy$x;y<-xy$y;x_marks<-xy$x_marks;y_marks<-xy$y_marks;deltas2_corr<-xy$deltas2_corr
  nx<-length(x[1,]);ny<-length(y[1,])
  x_marks_matrix<-rep(x_marks,ny)
  y_marks_matrix<-rep(y_marks,each=nx)
  densities_ratio_x<-evaluate_density(points=x,density_est=centre_density)
  densities_ratio_y<-evaluate_density(points=y,density_est=centre_density)
  densities_ratio<-  (rep(densities_ratio_x,ny)*x_marks_matrix + rep(densities_ratio_y,each=nx)*y_marks_matrix)/(x_marks_matrix+y_marks_matrix)/(
    rep(densities_ratio_x,ny)*rep(densities_ratio_y,each=nx))
  weights<-matrix(exp(-deltas2_corr/(sigma^2))*densities_ratio*c_vec[x_marks_matrix+y_marks_matrix]/
                    (c_vec[x_marks_matrix]*c_vec[y_marks_matrix]),nrow=nx,ncol=ny)
  return(weights)
}

#### OTHER AUXILIARY FUNCTIONS FOR THE METROPOLIS-WITHIN-GIBBS ALGORITHM ####
###
kd_cluster_count_output<-function(state,links_count,T,n_points,level=NULL){
  k<-length(n_points)-1
  n_points_each<-n_points[2:(k+1)]-n_points[1:k]
  if (length(links_count)==0){
    links_count<-simple_sparse_array(i=matrix(c(rep(1,k),n_points_each+1),nrow=2,ncol=k,byrow=TRUE),v=c(0,0))
  }#array(data=0,dim =(n_points_each+1))}
  if((length(level)==0 || level==1)&length(state$cluster)>0){
    cluster_indices<-state$cluster
    temporary<-state$cluster+n_points[1:k]
    cluster_indices[cluster_indices==0]<-temporary[cluster_indices==0]
    cluster_indices<-cluster_indices-n_points[1:k]+1
    links_count[t(cluster_indices)]<-links_count[t(cluster_indices)]+1/T
  }
  return(links_count)
}
evaluate_N_vec<-function(cluster,total_n_points){
  k<-dim(cluster)[1]
  if(length(cluster)==0){
    N<-c(total_n_points,rep(0,k-1))
  }else{
    cluster_sizes<-apply(cluster>0,2,sum)
    N<-as.vector(table(c(cluster_sizes,c(2:k))))-1
    N<-c(total_n_points-sum(N*c(2:k)),N)
  }
  return(N)
}
cut_density<-function(centre_density,n_grid,coast_distance=3000,wales_fraction_cutoff=1.9,all_fraction_cutoff=17){#wales_fraction_cutoff=1.6
  ####### FIND WHICH POINTS OF THE GRID FALL INSIDE THE UK
  grid_points<-SpatialPoints(coords=cbind(rep(centre_density$xcol,n_grid),rep(centre_density$yrow,each=n_grid)),
                             proj4string=CRS("+init=epsg:27700"))#CRS("+init=epsg:4326"))
  map_UK<-map('worldHires',c('UK','Isle of Wight'),fill=TRUE,plot=FALSE,xlim=c(-6,3), ylim=c(49,55))
  map_UK<-map2SpatialPolygons(map_UK,IDs=map_UK$names,proj4string=CRS("+init=epsg:4326"))#epsg:4326 is the latitude-longitude system
  UK_OSGB<-spTransform(map_UK,CRS("+init=epsg:27700"))#epsg:27700 is the national OSGB (British national grid system)
  indicator_vec<-over(grid_points,UK_OSGB)
  inside_vec<-which(indicator_vec>0)
  outside_vec<-(1:length(indicator_vec))[-inside_vec]
  inside_matrix<-matrix(rep(NA,n_grid^2),nrow=n_grid,ncol=n_grid)
  inside_matrix[inside_vec]<-TRUE
  inside_matrix[outside_vec]<-FALSE
  ######## FIND WHICH POINTS OF THE GRID FALL INSIDE WALES
  Wales_coordinates_open<-UK_OSGB@polygons[[2]]@Polygons[[1]]@coords[8460:10750,]
  end<-UK_OSGB@polygons[[2]]@Polygons[[1]]@coords[8460,]
  start<-UK_OSGB@polygons[[2]]@Polygons[[1]]@coords[10751,]
  xlink<-start[1]+(end[1]-start[1])*c(0:1000)/1000
  ylink<-start[2]+(end[2]-start[2])*c(0:1000)/1000
  Wales_coordinates<-rbind(Wales_coordinates_open,cbind(xlink,ylink))
  Wales.Polygon<-Polygon(coords=Wales_coordinates)
  Wales.Polygons<-Polygons(srl=list(Wales.Polygon),ID='Wales')
  Wales.SpatialPolygons<-SpatialPolygons(Srl=list(Wales.Polygons), proj4string=CRS("+init=epsg:27700"))
  indicator_vec_W<-over(grid_points,Wales.SpatialPolygons)
  inside_vec_W<-which(indicator_vec_W>0)
  outside_vec_W<-(1:length(indicator_vec_W))[-inside_vec_W]
  inside_matrix_W<-matrix(rep(NA,n_grid^2),nrow=n_grid,ncol=n_grid)
  inside_matrix_W[inside_vec_W]<-TRUE
  inside_matrix_W[outside_vec_W]<-FALSE
  ####### DO THE REGION CUTTING
  Wales_cutoff<-max(centre_density$v)/wales_fraction_cutoff
  all_cutoff<-max(centre_density$v)/all_fraction_cutoff
  matrix_selected<-inside_matrix&(!inside_matrix_W|(inside_matrix_W&t(centre_density$v)>Wales_cutoff))&(t(centre_density$v)>all_cutoff)
  ####### DO THE BORDER SMOOTING
  adjacient<-round(coast_distance/(centre_density$xcol[2]-centre_density$xcol[1]))
  matrix_selected_smoothed<-matrix_selected
  if (adjacient>0){
    for (i in 1:(2*adjacient)){
      for (j in 1:(2*adjacient)){
        matrix_selected_smoothed[adjacient:(n_grid-adjacient),adjacient:(n_grid-adjacient)]<-
          matrix_selected_smoothed[adjacient:(n_grid-adjacient),adjacient:(n_grid-adjacient)]|
          matrix_selected[i:(n_grid-2*adjacient+i),j:(n_grid-2*adjacient+j)]
      }
    }
  }
  centre_density_cs<-centre_density
  centre_density_cs$v[t(!matrix_selected_smoothed)]<-1e-100
  return(centre_density_cs)
}
evaluate_density<-function(points,density_est){
  points<-matrix(c(points),nrow=2)
  indices<-matrix(c(round((points[1,]-density_est$xmin)/(density_est$xmax-density_est$xmin)*(density_est$n_grid-1)+1),
                    round((points[2,]-density_est$ymin)/(density_est$ymax-density_est$ymin)*(density_est$n_grid-1)+1)),ncol=2)
  return(density_est$z[indices])
}
intracluster_sqrdist<-function(vec,n_points,points){
  k<-length(n_points)-1
  coord<-c()
  for (i in (1:k)){
    if(vec[i]>0){
      coord<-cbind(coord,points[[i]][,vec[i]-n_points[i]])
    }
  }
  return(sum((coord-apply(coord,1,mean))^2))
}